"""
Issue report module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import io
import re
import subprocess  # nosec B404 # nosemgrep: gitlab.bandit.B404
from tempfile import NamedTemporaryFile
from concurrent.futures import Future

from typing import TYPE_CHECKING, List, Optional
from gi.repository import Gtk, GLib

from proton.session.exceptions import ProtonAPINotReachable, ProtonAPIError
from proton.vpn.session.dataclasses import BugReportForm
from proton.vpn.app.gtk import __version__
from proton.vpn import logging
from proton.vpn.app.gtk.utils.executor import AsyncExecutor
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar

if TYPE_CHECKING:
    from proton.vpn.app.gtk.controller import Controller
    from proton.vpn.app.gtk.app import MainWindow

logger = logging.getLogger(__name__)


class BugReportDialog(Gtk.Dialog):  # pylint: disable=too-many-instance-attributes
    """Widget used to report bug/issues to Proton."""
    WIDTH = 400
    HEIGHT = 300
    EMAIL_REGEX = re.compile(
        r'[^@\s]+@[^@\s]{2,}\.[^@\s\.-]{2,}'
    )
    BUG_REPORT_SENDING_MESSAGE = "Reporting your issue..."
    BUG_REPORT_SUCCESS_MESSAGE = "Your issue has been reported"
    BUG_REPORT_NETWORK_ERROR_MESSAGE = "Proton services could not be reached.\n" \
                                       "Please try again."
    BUG_REPORT_UNEXPECTED_ERROR_MESSAGE = "Something went wrong. " \
                                          "Please try submitting your report at:\n" \
                                          "https://protonvpn.com/support-form"
    BUG_REPORT_TITLE = "Report from Linux app"
    BUG_REPORT_CLIENT = "Linux GUI"
    BUG_REPORT_VERSION = __version__
    BUG_REPORT_DESCRIPTION_MIN_CHARACTERS = 50

    def __init__(
        self, controller: Controller, main_window: MainWindow,
        notification_bar: Optional[NotificationBar] = None,
        log_collector: Optional["LogCollector"] = None
    ):
        super().__init__()
        self.set_name("bug-report-dialog")
        self._controller = controller
        self._main_window = main_window
        self.notification_bar = notification_bar or NotificationBar()
        self._log_collector = log_collector or LogCollector(
            self._controller.executor
        )

        self.set_title("Report an Issue")
        self.set_default_size(BugReportDialog.WIDTH, BugReportDialog.HEIGHT)

        self.cancel_button = self.add_button("_Cancel", Gtk.ResponseType.CANCEL)
        self.submit_button = self.add_button("_Submit", Gtk.ResponseType.OK)

        self.cancel_button.add_css_class("danger")
        self.submit_button.add_css_class("primary")

        safe_signal_connect(self, "response", self._on_response)

        self._generate_fields()
        self.submit_button.set_sensitive(False)

    @property
    def status_label(self) -> str:
        """Returns the current message of the notification bar."""
        return self.notification_bar.current_message

    def _on_response(self, _: BugReportDialog, response: Gtk.ResponseType):
        """Upon any of the button being clicked in the dialog,
        it's responde is evaluated.

        It first starts the background process to generate the logs and only after
        those are finished being generated, we instantiate `BugReportForm` will
        all the available data.
        """
        if response != Gtk.ResponseType.OK:
            self.close()
            return

        # Time here has to be long to account for network issues or when API is not
        # reacheable.
        self.notification_bar.show_info_message(self.BUG_REPORT_SENDING_MESSAGE, 60000)
        if self.send_logs_checkbox.get_active():
            logs_future = self._log_collector.get_logs()
            logs_future.add_done_callback(
                lambda _logs_future: GLib.idle_add(
                    self._submit_bug_report, _logs_future.result()
                )
            )
        else:
            GLib.idle_add(self._submit_bug_report, [])

        # Prevent that the window closes before receiving the API response,
        # as by default Gtk.Dialog closes after the response signal is emitted.
        # https://lazka.github.io/pgi-docs/#GObject-2.0/functions.html#GObject.signal_stop_emission_by_name
        self.stop_emission_by_name("response")

    def _submit_bug_report(self, logs: List[io.IOBase]):
        report_form = BugReportForm(
            username=self.username_entry.get_text(),
            email=self.email_entry.get_text(),
            title=self.BUG_REPORT_TITLE,
            description=self.description_buffer.get_text(
                self.description_buffer.get_start_iter(),
                self.description_buffer.get_end_iter(),
                True
            ),
            client_version=self.BUG_REPORT_VERSION,
            client=self.BUG_REPORT_CLIENT,
            attachments=logs
        )
        self._disable_form()
        future = self._controller.submit_bug_report(report_form)
        future.add_done_callback(
            lambda future: GLib.idle_add(
                self._on_report_submission_result,
                future, report_form
            )
        )

    def _on_report_submission_result(self, future: Future, report_form: BugReportForm):
        try:
            future.result()
        except ProtonAPINotReachable:
            logger.warning("Report submission failed: API not reachable.")
            self.notification_bar.show_error_message(
                self.BUG_REPORT_NETWORK_ERROR_MESSAGE
            )
            self._enable_form()
        except ProtonAPIError as exc:
            # ProtonAPIError is raised when the backend considers the email
            # address is not valid (some addresses like test@test.com are banned).
            logger.warning(f"Proton API error: {exc}")
            self.notification_bar.show_error_message(exc.error)
            self._enable_form()
        except Exception:  # pylint: disable=broad-except
            self.notification_bar.show_error_message(
                self.BUG_REPORT_UNEXPECTED_ERROR_MESSAGE
            )
            self._enable_form()
            logger.exception("Unexpected error submitting bug report.")
        else:
            self._main_window.main_widget.notifications.show_success_message(
                self.BUG_REPORT_SUCCESS_MESSAGE
            )
            self.close()
        finally:
            for attachment in report_form.attachments:
                attachment.close()

        return False

    def _disable_form(self):
        self.username_entry.set_sensitive(False)
        self.email_entry.set_sensitive(False)
        self.description_textview.set_sensitive(False)
        self.send_logs_checkbox.set_sensitive(False)
        self.submit_button.set_sensitive(False)

    def _enable_form(self):
        self.username_entry.set_sensitive(True)
        self.email_entry.set_sensitive(True)
        self.description_textview.set_sensitive(True)
        self.send_logs_checkbox.set_sensitive(True)
        if self._can_user_submit_form:
            self.submit_button.set_sensitive(True)

    def _on_entry_changed(self, _: Gtk.Widget):
        self.submit_button.set_sensitive(self._can_user_submit_form)

    @property
    def _can_user_submit_form(self) -> bool:
        is_username_provided = len(self.username_entry.get_text().strip()) > 0
        is_email_provided = re.fullmatch(
            self.EMAIL_REGEX, self.email_entry.get_text()
        )
        is_description_provided = len(self.description_buffer.get_text(
            self.description_buffer.get_start_iter(),
            self.description_buffer.get_end_iter(),
            True
        )) > BugReportDialog.BUG_REPORT_DESCRIPTION_MIN_CHARACTERS

        return bool(
            is_username_provided
            and is_email_provided
            and is_description_provided
        )

    def _generate_fields(self):  # pylint: disable=too-many-statements
        """Generates the necessary fields for the report."""
        layout = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
        layout.set_margin_top(0)
        layout.set_margin_bottom(0)
        layout.set_margin_start(0)
        layout.set_margin_end(0)
        layout.append(self.notification_bar)

        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=3)
        content.set_name("bug-report-content")
        layout.append(content)

        username_label = Gtk.Label(label="Username")
        username_label.set_halign(Gtk.Align.START)

        self.username_entry = Gtk.Entry()
        self.username_entry.set_property("margin-bottom", 10)
        self.username_entry.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
        self.username_entry.set_name("username")
        content.append(username_label)
        content.append(self.username_entry)

        email_label = Gtk.Label(label="Email")
        email_label.set_halign(Gtk.Align.START)

        self.email_entry = Gtk.Entry()
        self.email_entry.set_property("margin-bottom", 10)
        self.email_entry.set_input_purpose(Gtk.InputPurpose.EMAIL)
        self.email_entry.set_name("email")
        content.append(email_label)
        content.append(self.email_entry)

        min_characters = BugReportDialog.BUG_REPORT_DESCRIPTION_MIN_CHARACTERS
        description_label = Gtk.Label(
            label=f"Description (minimum {min_characters} characters)"
        )

        description_label.set_halign(Gtk.Align.START)
        # Has to have min 50 chars
        self.description_buffer = Gtk.TextBuffer()
        self.description_textview = Gtk.TextView.new_with_buffer(
            self.description_buffer
        )
        self.description_textview.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.description_textview.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
        self.description_textview.set_justification(Gtk.Justification.FILL)
        self.description_textview.set_name("description")
        scrolled_window_textview = Gtk.ScrolledWindow()
        scrolled_window_textview.set_property("margin-bottom", 10)
        scrolled_window_textview.set_min_content_height(100)
        scrolled_window_textview.set_child(self.description_textview)
        content.append(description_label)
        content.append(scrolled_window_textview)

        self.send_logs_checkbox = Gtk.CheckButton.new_with_label("Send error logs")
        self.send_logs_checkbox.set_active(True)
        self.send_logs_checkbox.set_name("send_logs")
        content.append(self.send_logs_checkbox)

        content_area = self.get_content_area()
        content_area.append(layout)
        content_area.set_margin_top(0)
        content_area.set_margin_bottom(0)
        content_area.set_margin_start(0)
        content_area.set_margin_end(0)
        content_area.set_spacing(20)

        safe_signal_connect(
            self.username_entry, "changed", self._on_entry_changed
        )
        safe_signal_connect(
            self.email_entry, "changed", self._on_entry_changed
        )
        safe_signal_connect(
            self.description_buffer, "changed", self._on_entry_changed
        )

    def get_submit_button(self):
        """Returns the Submit button."""
        return self.submit_button

    def click_on_submit_button(self):
        """Clicks the Submit button."""
        self.submit_button.emit("clicked")


class LogCollector:  # pylint: disable=too-few-public-methods
    """Collects all necessary logs needed for the report tool."""

    def __init__(self, executor: AsyncExecutor):
        self._executor = executor

    def get_logs(self) -> Future:
        """
        Generates and returns all available logs asynchronously.
        The future result is a List of file objects.
        """

        def _collect_logs():
            return [
                self._get_app_log(),
                self._generate_network_manager_log(),
                self._get_daemon_log()
            ]

        return self._executor.submit(_collect_logs)

    def _get_app_log(self) -> io.IOBase:
        """Get app log"""
        root_logger = logger.logger.root
        for handler in root_logger.handlers:
            if handler.__class__.__name__ == "RotatingFileHandler":
                return open(handler.baseFilename, "rb")

        raise RuntimeError("App logs not found.")

    def _get_daemon_log(self) -> io.IOBase:
        with NamedTemporaryFile(
            prefix="SplitTunneling-", suffix=".log", delete=False
        ) as temp_file:
            args = [
                "journalctl", "-u", "me.proton.vpn.split_tunneling", "--no-pager",
                "--utc", "--since=-1d", "--no-hostname"
            ]
            process = subprocess.run(args,  # nosec B603 # noqa E501 # pylint: disable=no-member, disable=line-too-long # nosemgrep: gitlab.bandit.B604, python.lang.security.audit.dangerous-subprocess-use-audit.dangerous-subprocess-use-audit
                                            stdout=temp_file,
                                            check=False)
            if process.returncode == 0:
                # We're waiting for the process to terminate before
                # reading the file, so we should be okay to open it.
                return open(temp_file.name, "rb")  # pylint: disable=line-too-long # noqa: E501 # nosemgrep: python.lang.correctness.tempfile.flush.tempfile-without-flush

            raise RuntimeError("Split tunneling logs could not be generated.")

    def _generate_network_manager_log(self) -> io.IOBase:
        """Generate Network Manager logs"""
        with NamedTemporaryFile(
            prefix="NetworkManager-", suffix=".log", delete=False
        ) as temp_file:
            args = [
                "journalctl", "-u", "NetworkManager", "--no-pager",
                "--utc", "--since=-1d", "--no-hostname"
            ]
            process = subprocess.run(args,  # nosec B603 # noqa E501 # pylint: disable=no-member, disable=line-too-long # nosemgrep: gitlab.bandit.B604, python.lang.security.audit.dangerous-subprocess-use-audit.dangerous-subprocess-use-audit
                                            stdout=temp_file,
                                            check=False)
            if process.returncode == 0:
                # We're waiting for the process to terminate before
                # reading the file, so we should be okay to open it.
                return open(temp_file.name, "rb")  # pylint: disable=line-too-long # noqa: E501 # nosemgrep: python.lang.correctness.tempfile.flush.tempfile-without-flush

            raise RuntimeError("Network Manager logs could not be generated.")



"""
General settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING, Optional, Callable
from gi.repository import Gtk, GLib, Gio
from proton.vpn import logging
from proton.vpn.connection import states
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, ToggleWidget, EntryWidget,
)
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.early_access import \
    EarlyAccessWidget
if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.main.tray_indicator import TrayIndicator

_PROTOCOL_GROUPS = ("generic", "protun")

logger = logging.getLogger(__name__)


class TrayPinnedServersWidget(EntryWidget):
    """Custom widget that holds the pinned servers to tray setting."""
    TRAY_PINNED_SERVERS_LABEL = "Pinned tray connections"
    TRAY_PINNED_SERVERS_DESCRIPTION = "Access preferred connections from system tray."\
        " Enter country or server codes, separated by commas, to quickly connect "\
        "(e.g.: NL#42, JP, US, IT#01)."
    SETTING_NAME = "app_configuration.tray_pinned_servers"

    def __init__(self, controller: Controller, tray_indicator: Optional["TrayIndicator"] = None):
        super().__init__(
            controller=controller,
            title=self.TRAY_PINNED_SERVERS_LABEL,
            description=self.TRAY_PINNED_SERVERS_DESCRIPTION,
            setting_name=self.SETTING_NAME,
            callback=self._save_and_reload_pinned_servers
        )
        self._controller = controller
        self._tray_indicator = tray_indicator

    def _save_and_reload_pinned_servers(self, entry: Gtk.Entry, *_):
        self.save_setting(entry.get_text())
        if self._tray_indicator:
            self._tray_indicator.reload_pinned_servers()

    def get_setting(self):
        """Shortcut property that sets the new setting and stores to disk."""
        tray_pinned_servers = self._controller.get_setting_attr(self.SETTING_NAME)
        return ', '.join(tray_pinned_servers)

    def save_setting(self, new_value: str):  # noqa: F811
        """Returns if the the upgrade tag has overridden original interactive
        object."""
        server_list = []

        for pinned_server in new_value.split(","):
            cleaned_pinned_server = pinned_server.strip().upper()

            if cleaned_pinned_server:
                server_list.append(cleaned_pinned_server)

        self._controller.save_setting_attr(self.SETTING_NAME, server_list)


def default_file_browser(packet_capture_widget: "PacketCaptureWidget"):
    """
    Returns a callback function for opening a file browser dialog to select
    the packet capture directory. The callback is only returned if
    Gtk.FileDialog is available (GTK >= 4.10); otherwise, returns None.
    """
    if not hasattr(Gtk, "FileDialog"):
        return None  # Gtk.FileDialog is only available in GTK >= 4.10

    def _on_browse_response(dialog, result):
        try:
            gfile = dialog.select_folder_finish(result)
            if gfile:
                path = gfile.get_path()
                if path:
                    packet_capture_widget.entry.set_text(path)
                    packet_capture_widget.save_setting(path)
        except GLib.Error as err:
            if err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                pass  # user cancelled
            else:
                raise

    def _on_browse_clicked(_button):
        # Guarded by the hasattr check in _build_ui; only reachable on GTK >= 4.10.
        dialog = Gtk.FileDialog.new()
        dialog.set_title("Select packet capture directory")
        dialog.set_accept_label("_Select")
        current_path = packet_capture_widget.entry.get_text().strip()
        if current_path:
            dialog.set_initial_folder(Gio.File.new_for_path(current_path))
        dialog.select_folder(packet_capture_widget.get_root(), None, _on_browse_response)

    return _on_browse_clicked


class PacketCaptureWidget(EntryWidget):
    """Packet capture settings row: start/stop toggle, file path entry, and browse button."""
    LABEL = "Create a troubleshooting file"
    DESCRIPTION = (
        "Captures a specific issue you're facing when using Proton VPN "
        "to get help from our customer support team.\n\n"
        "Warning: This file is a recording of all your internet activity during "
        "the capture session."
    )

    SETTING_NAME = "settings.packet_capture.directory_path"

    def __init__(self,
                 controller: Controller,
                 file_browser:
                     Callable[["PacketCaptureWidget"],
                              Optional[Callable]] = default_file_browser):
        self._on_browse_clicked = file_browser(self)

        super().__init__(controller=controller, title=self.LABEL,
                         description=self.DESCRIPTION,
                         setting_name=self.SETTING_NAME,
                         callback=self._on_focus_outside_entry)
        self._capturing = False
        self.connect("realize", self._on_realize)
        self.connect("unrealize", self._on_unrealize)

    def _on_focus_outside_entry(self, entry: Gtk.Entry, *_):
        self.save_setting(entry.get_text())

    def _build_ui(self):
        # Row 0: setting label + start/stop button (right-aligned, like other toggles)
        self.attach(self.label, 0, 0, 1, 1)

        self._start_stop_button = Gtk.Button(label="Start")
        self._start_stop_button.set_sensitive(self._controller.is_connection_active)
        self._start_stop_button.connect("clicked", self._on_start_stop_clicked)
        self._start_stop_button.set_hexpand(True)
        self._start_stop_button.set_halign(Gtk.Align.END)
        self.attach(self._start_stop_button, 1, 0, 1, 1)

        # Row 1: description
        if self.description:
            self.attach(self.description, 0, 1, 2, 1)

        # Row 2: directory path entry + browse button spanning both columns
        file_row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        self.entry.set_hexpand(True)
        file_row.append(self.entry)

        if self._on_browse_clicked is not None:
            browse_button = Gtk.Button(label="Browse…")
            browse_button.connect("clicked", self._on_browse_clicked)
            file_row.append(browse_button)

        file_row.set_hexpand(True)
        self.attach(file_row, 0, 2, 2, 1)

    def _on_realize(self, _widget):
        self._controller.register_connection_status_subscriber(self)

    def _on_unrealize(self, _widget):
        if self._capturing:
            self._controller.executor.submit(
                self._controller.current_connection.stop_packet_capture
            ).add_done_callback(lambda f: GLib.idle_add(self._set_capturing, False))
        self._controller.unregister_connection_status_subscriber(self)

    def status_update(self, connection_state):
        """Called from the async thread when the VPN connection state changes."""
        GLib.idle_add(self._on_connection_state_changed, connection_state)

    def _on_connection_state_changed(self, connection_state):
        is_connected = isinstance(connection_state, states.Connected)
        self._start_stop_button.set_sensitive(is_connected)
        if not is_connected and self._capturing:
            self._set_capturing(False)

    def _on_start_stop_clicked(self, _button):
        if self._capturing:
            self._controller.executor.submit(
                self._controller.current_connection.stop_packet_capture
            ).add_done_callback(lambda f: GLib.idle_add(self._set_capturing, False))
        else:
            self._controller.executor.submit(
                self._controller.current_connection.start_packet_capture
            ).add_done_callback(lambda f: GLib.idle_add(self._set_capturing, True))

    def _set_capturing(self, capturing: bool):
        self._capturing = capturing

        self._start_stop_button.set_label("Stop" if capturing else "Start")
        if capturing:
            self._start_stop_button.add_css_class("destructive-action")
        else:
            self._start_stop_button.remove_css_class("destructive-action")

    @property
    def capturing(self) -> bool:
        """Returns whether packet capture is currently active."""
        return self._capturing


class GeneralSettings(BaseCategoryContainer):  # pylint: disable=too-many-instance-attributes
    """General settings are grouped under this class."""
    CATEGORY_NAME = "General"
    CONNECT_AT_APP_STARTUP_LABEL = "Auto connect"
    CONNECT_AT_APP_STARTUP_DESCRIPTION = "You will be connected to a server as "\
        "soon as Proton VPN app starts. Replace it with a country ISO code "\
        "(e.g.: US for United States), a server (e.g.: NL#42)"\
        " or Fastest for quick connection. Default value: Off."
    START_APP_MINIMIZED_LABEL = "Start app minimized"
    START_APP_MINIMIZED_DESCRIPTION = "When enabled, the app starts minimized "\
        "to the tray."
    ANONYMOUS_CRASH_REPORTS_LABEL = "Share anonymous crash reports"
    ANONYMOUS_CRASH_REPORTS_DESCRIPTION = "Crash reports help us fix bugs, detect firewalls, "\
        "and avoid VPN blocks.\n\nThese statistics do not contain your IP address, and they "\
        "cannot be used to identify you. We'll never share them with third parties."

    def __init__(
        self, controller: Controller,
        tray_indicator: Optional["TrayIndicator"] = None,
    ):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller
        self._tray_indicator = tray_indicator
        self._connect_at_startup_entry = None
        self._start_app_minimized_toggle = None
        self._anonymous_crash_reports_toggle = None
        self._packet_capture_widget: Optional[PacketCaptureWidget] = None

    def on_settings_changed(self, settings):
        """Update packet capture widget visibility when the protocol changes."""
        if self._packet_capture_widget is not None:
            self._packet_capture_widget.set_visible(
                self._protocol_supports_packet_capture(settings.protocol)
            )

    def _protocol_supports_packet_capture(self, protocol: str) -> bool:
        for group in _PROTOCOL_GROUPS:
            for cls in self._controller.get_available_protocols(group):
                if str(cls.protocol) == protocol:
                    return cls.supports_packet_capture()
        return False

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.build_connect_at_app_startup()

        if self._tray_indicator:
            self.build_start_app_minimized()
            self.build_tray_pinned_servers()

        self.build_anonymous_crash_reports()
        self.build_packet_capture()
        self.build_beta_upgrade()

    def build_connect_at_app_startup(self):
        """Builds and adds the `connect_at_app_startup` setting to the widget."""
        def _format_and_save_autoconnect_field(entry: Gtk.Entry, entry_widget: EntryWidget, *_):
            new_value: Optional[str] = entry.get_text().strip().upper()
            if new_value == "OFF":
                new_value = None

            entry_widget.save_setting(new_value)

        self._connect_at_startup_entry = EntryWidget(
            controller=self._controller,
            title=self.CONNECT_AT_APP_STARTUP_LABEL,
            description=self.CONNECT_AT_APP_STARTUP_DESCRIPTION,
            setting_name="app_configuration.connect_at_app_startup",
            callback=_format_and_save_autoconnect_field
        )
        self.append(self._connect_at_startup_entry)

    def build_start_app_minimized(self):
        """Builds and adds the `start_app_minimized` setting to the widget."""
        self._start_app_minimized_toggle = ToggleWidget(
            controller=self._controller,
            title=self.START_APP_MINIMIZED_LABEL,
            description=self.START_APP_MINIMIZED_DESCRIPTION,
            setting_name="app_configuration.start_app_minimized"
        )
        self.append(self._start_app_minimized_toggle)

    def build_tray_pinned_servers(self):
        """Builds and adds the `tray_pinned_servers` setting to the widget."""
        self.append(TrayPinnedServersWidget(
            controller=self._controller, tray_indicator=self._tray_indicator
        ))

    def build_anonymous_crash_reports(self):
        """Builds and adds the `anonymous_crash_reports` setting to the widget."""
        self._anonymous_crash_reports_toggle = ToggleWidget(
            controller=self._controller,
            title=self.ANONYMOUS_CRASH_REPORTS_LABEL,
            description=self.ANONYMOUS_CRASH_REPORTS_DESCRIPTION,
            setting_name="settings.anonymous_crash_reports"
        )
        self.append(self._anonymous_crash_reports_toggle)

    def build_packet_capture(self):
        """Builds and adds the `packet_capture` file path setting to the widget."""
        widget = PacketCaptureWidget(self._controller)
        current_protocol = self._controller.get_settings().protocol
        widget.set_visible(self._protocol_supports_packet_capture(current_protocol))
        self._packet_capture_widget = widget
        self.append(widget)

    def build_beta_upgrade(self):
        """Builds and adds the `Early Access` setting to the widget."""
        early_access = EarlyAccessWidget(self._controller)

        if not early_access.can_early_access_be_displayed():
            return

        self.append(early_access)



"""
Account settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from gi.repository import Gio
from proton.vpn import logging
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, CustomButton
)

logger = logging.getLogger(__name__)


class AccountSettings(BaseCategoryContainer):  # pylint: disable=too-many-instance-attributes
    """Account settings are grouped under this class."""
    CATEGORY_NAME = "Account"
    MANAGE_ACCOUNT_URL = "https://account.protonvpn.com/account"

    def __init__(self, controller: Controller):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.append(CustomButton(
            title=self._controller.account_name,
            description=f"VPN plan: {self._controller.account_data.plan_title or 'Free'}",
            button_label="Manage Account",
            on_click_callback=self._on_click_manage_account_button,
            bold_title=True
        ))

    def _on_click_manage_account_button(self, *_):
        Gio.AppInfo.launch_default_for_uri(self.MANAGE_ACCOUNT_URL, None)



"""
This module defines the menu that shown in the header bar.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING
from concurrent.futures import Future

from gi.repository import Gio, GLib, GObject
from proton.vpn.app.gtk import Gtk

from proton.vpn.connection.states import State, Disconnected
from proton.vpn.app.gtk.widgets.headerbar.menu.bug_report_dialog import BugReportDialog
from proton.vpn.app.gtk.widgets.headerbar.menu.about_dialog import AboutDialog
from proton.vpn.app.gtk.widgets.main.confirmation_dialog import ConfirmationDialog
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget, DefaultLoadingWidget
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings import SettingsWindow
from proton.vpn.app.gtk.widgets.headerbar.menu.release_notes_dialog import ReleaseNotesDialog
from proton.vpn.connection.enum import KillSwitchSetting as KillSwitchSettingEnum

from proton.session.exceptions import ProtonAPINotReachable
from proton.vpn import logging

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from proton.vpn.app.gtk.app import MainWindow


class Menu(Gio.Menu):  # pylint: disable=too-many-instance-attributes
    """App menu shown in the header bar."""

    LOGOUT_LOADING_MESSAGE = "Signing out..."
    UNABLE_TO_LOGOUT_MESSAGE = "Unable to sign out, please ensure you have internet access."
    DISCONNECT_ON_LOGOUT_MESSAGE = "Signing out will cancel the current VPN connection.\n\n" \
                                   "Do you want to continue?"
    DISCONNECT_ON_LOGOUT_WITH_KILL_SWITCH_ENABLED_MESSAGE = "Signing out "\
        "will cancel the current VPN connection and disable the kill switch."\
        "\n\nDo you want to continue?"
    LOGOUT_AND_KILL_SWITCH_ENABLED_MESSAGE = "Signing out will "\
        "disable the kill switch, potentially exposing your internet traffic. "\
        "\n\nDo you want to continue?"
    DISCONNECT_ON_QUIT_MESSAGE = "Quitting the application will cancel the current" \
                                 " VPN connection.\n\nDo you want to continue?"
    DISCONNECT_ON_QUIT_WITH_PERMANENT_KILL_SWITCH_ENABLED_MESSAGE = "Quitting the application "\
        "will keep the kill switch active, but your current VPN connection will be terminated."\
        "\n\nDo you want to continue?"
    DISCONNECT_TITLE = "Active connection found"
    KILLSWITCH_ENABLED_TITLE = "Kill Switch enabled"

    def __init__(
        self, controller: Controller,
        main_window: "MainWindow", overlay_widget: OverlayWidget
    ):
        super().__init__()
        self._main_window = main_window
        self._controller = controller
        self._overlay_widget = overlay_widget

        self.bug_report_action = Gio.SimpleAction.new("report", None)
        self.settings_action = Gio.SimpleAction.new("settings", None)
        self.release_notes_action = Gio.SimpleAction.new("release_notes", None)
        self.about_action = Gio.SimpleAction.new("about", None)
        self.logout_action = Gio.SimpleAction.new("logout", None)
        self.quit_action = Gio.SimpleAction.new("quit", None)

        self.append_item(Gio.MenuItem.new("About", "win.about"))
        self.append_item(Gio.MenuItem.new("Settings", "win.settings"))
        self.append_item(Gio.MenuItem.new("Release notes", "win.release_notes"))
        self.append_item(Gio.MenuItem.new("Report an issue", "win.report"))
        self.append_item(Gio.MenuItem.new("Sign out", "win.logout"))
        self.append_item(Gio.MenuItem.new("Quit", "win.quit"))

        self._settings_window = None
        self._bug_dialog = None
        self._release_notes = None
        self._dialog_callback = None

        self._setup_actions()

    def status_update(self, connection_status: State):
        """
        This method is set as a callback to monitor the VPN connection after
        the user clicks on the quit menu option.
        """
        if isinstance(connection_status, Disconnected):
            self._main_window.quit()

    @property
    def logout_enabled(self) -> bool:
        """Returns if logout button is enabled or disabled."""
        return self.logout_action.get_enabled()

    @logout_enabled.setter
    def logout_enabled(self, newvalue: bool):
        """Set the logout button to either be enabled or disabled."""
        self.logout_action.set_enabled(newvalue)

    @property
    def settings_enabled(self) -> bool:
        """Returns if logout button is enabled or disabled."""
        return self.settings_action.get_enabled()

    @settings_enabled.setter
    def settings_enabled(self, newvalue: bool):
        """Set the settings button to either be enabled or disabled."""
        self.settings_action.set_enabled(newvalue)

    @GObject.Signal
    def user_logged_out(self):
        """Signal emitted after a successful logout."""

    def close_settings_window(self):
        """Closes the settings window if it's open."""
        if self._settings_window:
            self._settings_window.close()

    def _setup_actions(self):
        # Add actions to Gtk.ApplicationWindow
        self._main_window.add_action(self.bug_report_action)
        self._main_window.add_action(self.settings_action)
        self._main_window.add_action(self.release_notes_action)
        self._main_window.add_action(self.about_action)
        self._main_window.add_action(self.logout_action)
        self._main_window.add_action(self.quit_action)

        # Connect actions to callbacks
        safe_signal_connect(
            self.bug_report_action, "activate", self._on_report_an_issue_clicked
        )
        safe_signal_connect(
            self.settings_action, "activate", self._on_settings_clicked
        )
        safe_signal_connect(
            self.release_notes_action, "activate", self._on_release_notes_clicked
        )
        safe_signal_connect(
            self.about_action, "activate", self._on_about_clicked
        )
        safe_signal_connect(
            self.logout_action, "activate", self._on_logout_clicked
        )
        safe_signal_connect(
            self.quit_action, "activate", self._on_quit_clicked
        )

    def _on_report_an_issue_clicked(self, *_):
        self._bug_dialog = BugReportDialog(self._controller, self._main_window)
        self._bug_dialog.set_transient_for(self._main_window)
        self._bug_dialog.set_modal(True)
        safe_signal_connect(self._bug_dialog, "unrealize", self._on_bug_dialog_unrealize)
        self._bug_dialog.present()

    def _on_bug_dialog_unrealize(self, _):
        self._bug_dialog = None

    def _on_settings_clicked(self,  *_):
        self._settings_window = SettingsWindow(
            self._controller,
            self._main_window.application.tray_indicator
        )
        self._settings_window.set_transient_for(self._main_window)
        safe_signal_connect(self._settings_window, "unrealize", self._on_unrealize)
        self._settings_window.present()

    def _on_unrealize(self, _):
        self._settings_window = None

    def _on_release_notes_clicked(self,  *_):
        self._release_notes = ReleaseNotesDialog()
        self._release_notes.set_transient_for(self._main_window)
        safe_signal_connect(self._release_notes, "unrealize", self._on_release_notes_unrealize)
        self._release_notes.present()

    def _on_release_notes_unrealize(self, _):
        self._release_notes = None

    def _on_about_clicked(self, *_):
        about_dialog = AboutDialog()
        about_dialog.set_transient_for(self._main_window)
        about_dialog.set_modal(True)
        about_dialog.present()

    def _on_logout_clicked(self, *_):
        logger.info("Logout button clicked", category="ui", subcategory="logout", event="click")

        self.logout_enabled = False
        kill_switch_state = self._controller.get_settings().killswitch

        def on_logout_confirmed(confirmed: bool):
            if confirmed:
                logger.info("Yes", category="ui", subcategory="dialog", event="logout")

                self._overlay_widget.show(DefaultLoadingWidget(self.LOGOUT_LOADING_MESSAGE))

                if kill_switch_state > KillSwitchSettingEnum.OFF:
                    future = self._controller.disable_killswitch()
                    future.add_done_callback(
                        lambda f: GLib.idle_add(self._on_killswitch_disabled_logout, f)
                    )
                    return
                self._request_logout()

        if not self._controller.connection_disconnected:
            dialog = ConfirmationDialog(
                self.DISCONNECT_ON_LOGOUT_MESSAGE
                if kill_switch_state < KillSwitchSettingEnum.ON
                else self.DISCONNECT_ON_LOGOUT_WITH_KILL_SWITCH_ENABLED_MESSAGE,
                self.DISCONNECT_TITLE
            )
            self._display_dialog(dialog, on_logout_confirmed)
        elif kill_switch_state == KillSwitchSettingEnum.PERMANENT:
            self._display_dialog(
                ConfirmationDialog(
                    self.LOGOUT_AND_KILL_SWITCH_ENABLED_MESSAGE,
                    self.KILLSWITCH_ENABLED_TITLE
                ),
                on_logout_confirmed
            )
        else:
            on_logout_confirmed(True)

    def _on_quit_clicked(self, *_):
        kill_switch_state = self._controller.get_settings().killswitch

        if self._controller.connection_disconnected:
            self._main_window.quit()
        else:
            def on_quit_confirmed(confirmed: bool):
                if confirmed:
                    logger.info("Yes", category="ui", subcategory="dialog", event="quit")
                    self._controller.register_connection_status_subscriber(self)
                    future = self._controller.disconnect()
                    future.add_done_callback(lambda f: GLib.idle_add(f.result))
            dialog = ConfirmationDialog(
                self.DISCONNECT_ON_QUIT_WITH_PERMANENT_KILL_SWITCH_ENABLED_MESSAGE
                if kill_switch_state == KillSwitchSettingEnum.PERMANENT
                else self.DISCONNECT_ON_QUIT_MESSAGE,
                self.DISCONNECT_TITLE
            )
            self._display_dialog(dialog, on_quit_confirmed)

    def _on_killswitch_disabled_logout(self, future: Future):
        future.result()
        self._request_logout()

    def _request_logout(self):
        future = self._controller.logout()
        future.add_done_callback(
            lambda future: GLib.idle_add(self._on_logout_result, future)
        )

    def _on_logout_result(self, future: Future):
        """Callback when attempting to log out.
        Mainly used to emit if a successful logout has happened, or if a
            connection is found at logout, to display the dialog to the user.
        """
        try:
            future.result()
            logger.info(
                "Successful logout",
                category="app", subcategory="logout", event="success"
            )
            self.emit("user-logged-out")
        except ProtonAPINotReachable as e:  # pylint: disable=invalid-name
            logger.info(
                getattr(e, 'message', repr(e)),
                category="app", subcategory="logout", event="fail"
            )
            self._main_window.main_widget.notifications.show_error_message(
                self.UNABLE_TO_LOGOUT_MESSAGE
            )
            self.logout_enabled = True
        finally:
            self._overlay_widget.hide()

    def _display_dialog(self, dialog: ConfirmationDialog, callback):
        dialog.set_transient_for(self._main_window)
        dialog.set_modal(True)

        self._dialog_callback = callback
        safe_signal_connect(dialog, "response", self._on_dialog_response)
        dialog.present()

    def _on_dialog_response(self, dialog, response_id):
        dialog.destroy()
        self.logout_enabled = response_id in (
            Gtk.ResponseType.NO, Gtk.ResponseType.DELETE_EVENT
        )
        result = response_id == Gtk.ResponseType.YES

        if self._dialog_callback is not None:
            self._dialog_callback(result)
            self._dialog_callback = None

    def bug_report_button_click(self):
        """Clicks the bug report menu entry."""
        self._on_report_an_issue_clicked(self.bug_report_action)

    def about_button_click(self):
        """Clicks the about menu entry."""
        self._on_about_clicked(self.about_action)

    def logout_button_click(self):
        """Clicks the logout menu entry."""
        self._on_logout_clicked(self.logout_action)

    def quit_button_click(self):
        """Clicks the quit menu entry."""
        self._on_quit_clicked(self.quit_action)



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import Optional
from gi.repository import Gtk, GObject

from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.data_structures \
    import AppRowWithRemoveButton, AppData


class SelectedAppList(Gtk.ScrolledWindow):
    """A scrolled window that contains the list of split tunneling apps.

    It modifies the UI based on the provided input. It does not save any
    data to disk.
    """
    MAX_AMOUNT_OF_APPS_TO_DISPLAY = 6
    APP_ROW_SIZE = 35

    def __init__(self, apps_to_add: Optional[list[AppData]] = None):
        super().__init__()
        self.set_name("selected-app-list")

        self.main_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self._data_mapping: dict[AppData, AppRowWithRemoveButton] = {}

        viewport = Gtk.Viewport()
        viewport.add_css_class("viewport-frame")
        viewport.set_child(self.main_container)

        self.set_propagate_natural_width(True)
        self.set_child(viewport)

        self.refresh(apps_to_add)

    def refresh(self, selected_apps: list[AppData]):
        """Refreshes the entire list.
        This method is mainly used when adding new apps,
        so that the entire list can be refreshed.

        Args:
            selected_apps (list[AppData])
        """
        list_has_been_modified = False

        for app_data in list(self._data_mapping.keys()):
            # This app was already previously selected, we can skip it
            if app_data in selected_apps:
                continue

            # App was deselected, has to be removed
            list_has_been_modified = True
            self.main_container.remove(self._data_mapping.pop(app_data))

        for app_data in selected_apps:
            # If selected app exists in existing_apps then we don't add it
            if app_data in self._data_mapping:
                continue

            # Only add if child does not exist already
            list_has_been_modified = True
            app_row = self._generate_app_row(app_data)
            self.main_container.append(app_row)
            self._data_mapping[app_data] = app_row

        # Resize view only if list has been modified
        if not list_has_been_modified:
            return

        self._update_scrolled_window_size()
        self.emit("app-list-refreshed", selected_apps)

    def _remove_app(self, app_row: AppRowWithRemoveButton):
        """Remove app from list and notify once it's removed.

        Args:
            app_row (AppData)
        """
        app_data = app_row.app_data
        self.main_container.remove(self._data_mapping.pop(app_data))
        self._update_scrolled_window_size()
        self.emit("app-removed", app_data)

    def _update_scrolled_window_size(self):
        number_of_apps = len(self._data_mapping)
        no_scroll = number_of_apps < self.MAX_AMOUNT_OF_APPS_TO_DISPLAY
        max_size = self.MAX_AMOUNT_OF_APPS_TO_DISPLAY * self.APP_ROW_SIZE
        current_app_based_size = number_of_apps * self.APP_ROW_SIZE

        self.set_propagate_natural_height(no_scroll)
        self.set_min_content_height(
            max_size if current_app_based_size > max_size else current_app_based_size
        )

    def _generate_app_row(self, app_data: AppData) -> AppRowWithRemoveButton:
        app_row = AppRowWithRemoveButton.build(app_data)
        safe_signal_connect(app_row, "remove-app", self._remove_app)
        return app_row

    @GObject.Signal(name="app-removed", arg_types=(object,))
    def app_removed(self, app_data: AppData):
        """Signal emitted when an app is removed from the list.

        Args:
          app(AppData)
        """

    @GObject.Signal(name="app-list-refreshed", arg_types=[object,])
    def app_list_refreshed(self, selected_app_data: list[AppData]):
        """Signal emitted after the list has been fully refreshed.

        Args:
          app(AppData)
        """



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from types import ModuleType
from typing import Optional, Union

from dataclasses import dataclass, asdict

from gi.repository import Gtk, GObject, GdkPixbuf, GLib, Gdk


from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import \
    SettingName
from proton.vpn.app.gtk.assets import icons


ICON_SIZE_IN_PX = 16


@dataclass(eq=True, frozen=True)
class AppData:  # pylint: disable=missing-class-docstring
    name: str
    executable: str
    icon_name: str

    def to_dict(self) -> dict[str, Union[str, bool]]:
        """Convert dataclass to dict

        Returns:
            dict[str, Union[str, bool]]
        """
        return asdict(self)

    @staticmethod
    def from_dict(data: dict[str, Union[str, bool]]) -> AppData:
        """Build object based on dict.

        Args:
            data (dict[str, Union[str, bool]])

        Returns:
            AppData
        """
        return AppData(
            name=data["name"],
            executable=data["executable"],
            icon_name=data["icon_name"],
        )


def _get_missing_icon_pixbuff() -> GdkPixbuf.Pixbuf:
    return icons.get("no-app-icon.svg")


def get_icon(img_path: Optional[str], gtk: ModuleType = Gtk) -> Gtk.Image:
    """Returns a Gtk.Image based either on the app path image or else
    uses a default one.
    """
    pixbuff = None

    # If it starts with / then we've received a path to an image
    if img_path:
        try:
            if img_path.startswith("/"):
                pixbuff = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                    filename=img_path,
                    width=ICON_SIZE_IN_PX,
                    height=-1,
                    preserve_aspect_ratio=True
                )
            else:
                display = Gdk.Display.get_default()
                theme = Gtk.IconTheme.get_for_display(display)
                icon_paintable = theme.lookup_icon(
                    icon_name=img_path,
                    fallbacks=None,
                    size=ICON_SIZE_IN_PX,
                    scale=1,
                    direction=Gtk.TextDirection.NONE,
                    flags=0
                )
                if icon_paintable:
                    return gtk.Image.new_from_paintable(icon_paintable)
        except GLib.Error:
            pass

    if not pixbuff:
        pixbuff = _get_missing_icon_pixbuff()

    texture = Gdk.Texture.new_for_pixbuf(pixbuff)
    return gtk.Image.new_from_paintable(texture)


class AppRowWithCheckbox(Gtk.Grid):
    """App row used to display when selecting apps to be split tunneled.
    """
    def __init__(
        self,
        app_data: AppData,
        checked: bool,
        gtk: ModuleType = Gtk
    ):
        super().__init__()
        self.set_column_spacing(10)
        self.set_name(app_data.executable)

        self.app_data = app_data
        self._checked = checked
        self.gtk = gtk
        self._check_button: Optional[Gtk.CheckButton] = None

    @staticmethod
    def build(app_data: AppData, checked: bool) -> AppRowWithCheckbox:
        """Method to automate building the object

        Args:
            app_data (AppData):
            checked (bool)

        Returns:
            AppRowWithCheckbox
        """
        app_row = AppRowWithCheckbox(app_data, checked)
        app_row.build_ui()

        return app_row

    def build_ui(self):
        """Build the UI.
        It's still the parents responsibility to display it.
        """
        icon = get_icon(self.app_data.icon_name, self.gtk)
        label = SettingName(self.app_data.name)

        self._check_button = self.gtk.CheckButton.new()
        self._check_button.set_active(self._checked)

        self.attach(self._check_button, 0, 0, 1, 1)
        self.attach(icon, 1, 0, 1, 1)
        self.attach(label, 2, 0, 1, 1)

    @property
    def checked(self) -> bool:
        """Only available when built with `AppRow.build_with_remove_button()`.

        Returns:
            bool
        """
        return self._check_button.get_active()

    def _set_check(self, val: bool):
        """Mainly for testing purposes and not for public API.

        Args:
            val (bool)
        """
        self._check_button.set_active(val)


class AppRowWithRemoveButton(Gtk.Grid):
    """_summary_
    """
    def __init__(self, app_data: AppData, gtk: ModuleType = Gtk):
        super().__init__()
        self.set_column_spacing(10)
        self.set_name(app_data.executable)

        self.app_data = app_data
        self.gtk = gtk
        self._remove_button = None

    @staticmethod
    def build(app_data: AppData) -> AppRowWithRemoveButton:
        """Method to automate building the object

        Args:
            app_data (AppData)

        Returns:
            AppRowWithRemoveButton
        """
        app_row = AppRowWithRemoveButton(app_data)
        app_row.build_ui()

        return app_row

    def build_ui(self):
        """Build the UI.
        It's still the parents responsibility to display it.
        """
        icon = get_icon(self.app_data.icon_name, self.gtk)
        label = SettingName(self.app_data.name)
        self._remove_button = Gtk.Button.new_from_icon_name("edit-delete-symbolic")
        safe_signal_connect(self._remove_button, "clicked", self._signal_remove_app)

        self.attach(icon, 0, 0, 1, 1)
        self.attach(label, 1, 0, 1, 1)
        self.attach(self._remove_button, 2, 0, 1, 1)

    def _signal_remove_app(self, _: Gtk.Button):
        self.emit("remove-app")

    @GObject.Signal(name="remove-app")
    def remove_app(self):
        """
        Signal emitted after the user clicks on remove button.

        Since this object returns itself, we don't need to pass any arguments
        as we can easily access the `app` property.
        """

    def _click_on_remove_button(self):
        """Mainly for testing purposes and not for public API.
        """
        self._remove_button.emit("clicked")



"""
This module defines the main App class.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any, Optional
from gi.repository import GObject, Gtk, Gdk, GLib

from proton.vpn import logging

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.main.main_window import MainWindow
from proton.vpn.app.gtk.assets.style import STYLE_PATH
from proton.vpn.app.gtk.util import APPLICATION_ID, log_proton_package_versions
from proton.vpn.app.gtk.widgets.main.tray_indicator import TrayIndicator, TrayIndicatorNotSupported

logger = logging.getLogger(__name__)


class App(Gtk.Application):
    """
    Proton VPN GTK application.

    It inherits a set of common app functionality from Gtk.Application:
    https://docs.gtk.org/gtk3/class.Application.html.

    For example:
     - It guarantees that only one instance of the application is
       allowed (new app instances exit immediately if there is already
       an instance running).
     - It manages the windows associated to the application. The application
       exits automatically when the last one is closed.
     - It allows desktop shell integration by exporting actions and menus.
    """
    def __init__(
            self,
            controller: Controller
    ):
        super().__init__(application_id=APPLICATION_ID)
        logger.info(f"{self=}", category="APP", event="PROCESS_START")
        log_proton_package_versions()
        self._controller = controller
        self.window: Optional[MainWindow] = None
        self._tray_indicator = None
        self._signal_connect_queue: list[Any] = []
        self._start_minimized_from_cli = False
        self.add_options()

    def do_startup(self):  # pylint: disable=arguments-differ
        """Default GTK method.

        Runs at application startup, to load
        any necessary UI elements.
        """
        Gtk.Application.do_startup(self)
        css_provider = Gtk.CssProvider()
        css_provider.load_from_path(str(STYLE_PATH / "main.css"))

        display = Gdk.Display.get_default()
        Gtk.StyleContext.add_provider_for_display(
            display,
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def do_activate(self):  # pylint: disable=W0221
        """
        Method called by Gtk.Application when the default first window should
        be shown to the user.
        """
        if not self.window:
            self.window = MainWindow(self, self._controller)
            # Windows are associated with the application like this.
            # When the last one is closed, the application shuts down.
            self.add_window(self.window)
            # The behaviour of the button to close the window is configured
            # depending on whether the tray indicator is shown or not.
            self.window.configure_close_button_behaviour(
                tray_indicator_enabled=(self.tray_indicator is not None)
            )
            self.window.set_visible(True)

        self.window.present()
        self.emit("app-ready")

    def do_handle_local_options(self, options: GLib.VariantDict):  # noqa: E501 pylint: disable=arguments-differ
        """
        Handles the options defined in add_options
        Returns:
            Any negative number: Start as usual
            Zero: Stop without error
            Any positive number: Stop with the number as error code
        """
        if options.contains("version"):
            print(self._controller.app_version)
            return 0

        if options.contains("start-minimized"):
            self._start_minimized_from_cli = True

        return -1

    @property
    def error_dialog(self) -> Gtk.MessageDialog:
        """
        Gives access to currently opened error message dialogs. This method
        was made available for testing purposes.
        """
        return self.window.main_widget.notifications.error_dialog

    @GObject.Signal(name="app-ready")
    def app_ready(self):
        """Signal emitted when the app is ready for interaction."""
        if self._start_app_minimized and self.tray_indicator:
            self.window.set_visible(False)

    @property
    def _start_app_minimized(self) -> bool:
        return self._start_minimized_from_cli \
            or self._controller.get_app_configuration().start_app_minimized

    def add_options(self):
        """Adds the --start-minimized and --version command line options"""
        self.add_main_option(
            "start-minimized",
            0,
            GLib.OptionFlags(0),
            GLib.OptionArg.NONE,
            "Start minimized in the system tray"
        )

        self.add_main_option(
            "version",
            ord('v'),
            GLib.OptionFlags(0),
            GLib.OptionArg.NONE,
            "Display the application's version"
        )

    @property
    def tray_indicator(self):
        """Gives access to the tray indicator if it's installed and enabled."""
        if self._tray_indicator:
            return self._tray_indicator

        try:
            tray_indicator = TrayIndicator(self._controller)
            tray_indicator.setup(self.window)
        except TrayIndicatorNotSupported as excp:
            logger.warning(str(excp))
        else:
            self._tray_indicator = tray_indicator

        return self._tray_indicator



"""
This module defines the server list widget.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import time
from typing import List, Optional
import logging
from unittest.mock import Mock

from gi.repository import GLib, GObject

from proton.vpn import logging as proton_logging
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.session.servers import ServerList, TierEnum
from proton.vpn.session.servers.server_list_fetcher import ServerListFetcher

from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.country_row import CountryRow
from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.server_list_header_row import (
    ServerListHeaderRow,
)
from proton.vpn.app.gtk.widgets.vpn.search_entry import SearchEntry

from proton.vpn.app.gtk.widgets.vpn.serverlist.city_view.utils import (
    sync_rows_with_model_items,
)

logger = proton_logging.getLogger(__name__)


class ServerListWidget(Gtk.ScrolledWindow):
    """Server list widget displaying countries, locations and their servers."""

    def __init__(self, controller: Controller, search_entry: SearchEntry | None = None):
        super().__init__()
        self._controller = controller
        self._user_tier: Optional[int] = None
        self._search_entry = search_entry

        self.set_policy(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC
        )
        self.set_propagate_natural_width(True)
        self.set_name("server-list-widget")
        self.set_overlay_scrolling(False)

        # pylint: disable=duplicate-code
        self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._container.set_name("server-list-widget-container")
        self._container.set_vexpand(True)
        self._container.set_spacing(5)
        self.set_child(self._container)

        self._header_row = ServerListHeaderRow()
        self._container.prepend(self._header_row)

        self._country_rows_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._country_rows_container.set_name("country-rows-container")
        self._country_rows_container.set_vexpand(True)
        self._country_rows_container.set_spacing(5)
        self._container.append(self._country_rows_container)

        self._country_rows: List[CountryRow] = []

    def display(self, user_tier: int, server_list: ServerList):
        """Builds and displays the server list."""
        self._user_tier = user_tier
        self._populate_countries(server_list)
        self._controller.set_server_list_updated_callback(self._on_server_list_update)
        self._controller.set_server_loads_updated_callback(self._on_server_loads_update)
        self.emit("ui-updated")

    def focus_on_entry(self, _widget, name_to_search: str) -> None:
        """Searches for an entry by name and either connects to it directly,
           or focuses on it."""
        # pylint: disable=duplicate-code
        # Server
        if "#" in name_to_search:
            future = self._controller.connect_to_server(name_to_search)
            future.add_done_callback(lambda f: GLib.idle_add(f.result))
            if self._search_entry:
                self._search_entry.grab_focus()
            return
        for country in self.country_rows:
            # Country
            if country.country_name.lower() == name_to_search.lower():
                country.grab_focus()
                return

            # Location
            for location in country.locations:
                if location.name.lower() == name_to_search.lower():
                    country.focus_on_location(location.name)
                    return

    @GObject.Signal(name="ui-updated")
    def ui_updated(self):
        """Signal emitted once the server list within the UI has been updated.
        Mainly used for test purposes."""

    def _populate_countries(self, server_list: ServerList):
        self._display_country_rows(server_list)

    @property
    def country_rows(self) -> List[CountryRow]:
        """Returns the list of country rows currently displayed."""
        return list(self._country_rows)

    def _remove_country_rows(self):
        while self._country_rows:
            row = self._country_rows.pop()
            row.reset()
            self._country_rows_container.remove(row)

    def _display_country_rows(self, server_list: ServerList):
        free_user = self._user_tier == TierEnum.FREE
        countries = server_list.group_by_country(
            group_by_location=True,
            include_free_servers=free_user
        )
        if free_user:
            # If the current user has a free account, sort the countries having
            # free servers first.
            countries.sort(key=lambda country: (0 if country.free else 1, country.name))

        # Collect expanded states before refresh (keyed by country code and child group name)
        expanded_countries = {row.country_code.lower(): row.expanded for row in self.country_rows}
        expanded_groups_per_country = {
            country_row.country_code.lower(): set(
                location_row.label.lower() for location_row in (
                    country_row.location_rows
                    + ([country_row.secure_core_row] if country_row.secure_core_row else [])
                )
                if location_row.expanded
            )
            for country_row in self.country_rows
        }

        def display_country_row(row, country):
            expanded = expanded_countries.get(country.code.lower(), False)
            expanded_groups = expanded_groups_per_country.get(country.code.lower())
            row.display(
                self._controller, country, self._user_tier,
                expanded=expanded, expanded_groups=expanded_groups
            )

        sync_rows_with_model_items(
            countries,
            self._country_rows,
            self._country_rows_container,
            CountryRow,
            display_country_row
        )
        self._header_row.set_count(len(countries))

    def _on_server_list_update(self):
        """Whenever a new server list is received the UI should be updated."""
        start = time.time()
        self.display(self._user_tier, self._controller.server_list)
        logger.info(
            "Full server list widget update completed in "
            f"{time.time() - start:.2f} seconds."
        )

    def _on_server_loads_update(self):
        start = time.time()
        self.display(self._user_tier, self._controller.server_list)
        logger.info(
            "Partial server list widget update completed in "
            f"{time.time() - start:.2f} seconds."
        )

    def unload(self):
        """Unloads the server list widget and its resources."""
        self._controller.unset_server_list_updated_callback()
        self._controller.unset_server_loads_updated_callback()
        self._remove_country_rows()


def _on_activate(app):
    server_list_widget = ServerListWidget(controller=Mock(spec=Controller))

    win = Gtk.ApplicationWindow(application=app)
    win.set_default_size(400, 600)
    win.set_title("Server List")
    win.get_settings().props.gtk_application_prefer_dark_theme = True
    win.set_child(server_list_widget)
    _load_cached_server_list(server_list_widget)
    GLib.timeout_add_seconds(5, _load_cached_server_list, server_list_widget)
    win.present()


def _load_cached_server_list(server_list_widget: ServerListWidget):
    logger.info("Refreshing server list")
    server_list = ServerListFetcher(session=None).load_from_cache()
    server_list_widget.display(user_tier=2, server_list=server_list)
    return True


def main():
    """Main entry point for testing the server list widget standalone."""
    logger.setLevel(logging.DEBUG)
    app = Gtk.Application()
    safe_signal_connect(app, 'activate', _on_activate)

    app.run(None)


if __name__ == "__main__":
    main()



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations


from types import ModuleType
from typing import cast
from gi.repository import Gtk, GObject


from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app.data_structures \
    import AppRowWithCheckbox, AppData


class AppSelectionWindow(Gtk.Window):
    """This widget displays all available apps,
    and pre-check any already selected apps. Once the user
    clicks on `Done`, all data is transferred back to
    `AppBasedSplitTunnelingSettings` for further processing,
    both for UI updates and also storing data to disk.
    """
    def __init__(
        self,
        title: str,
        controller: Controller,
        stored_apps: list[str],
        installed_apps: list[AppData],
        gtk: ModuleType = Gtk
    ):  # pylint: disable=too-many-arguments
        super().__init__()
        self.set_modal(True)
        self.set_title(title)
        self.set_default_size(600, 500)
        self.set_name("split-tunneling-app-selection-window")

        self.gtk = gtk
        self._controller = controller
        self._stored_apps = stored_apps
        self._installed_apps = installed_apps

        # Can't use self.container as that seems to be a non-writable property,
        # probably reserved by Gtk.
        self.main_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.content_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10)

        self._create_elastic_window()

        safe_signal_connect(self, "realize", self._build_ui)

    def _create_elastic_window(self):
        """This allows for the content to be always centered and expand or contract
        based on window size.

        The reason we use two containers is mainly due to the notification bar, as this
        way the notification will span across the entire window while only the
        settings will be centered.
        """
        viewport = Gtk.Viewport.new(None, None)
        viewport.add_css_class("viewport-frame")
        viewport.set_child(self.content_container)

        scrolled_window = Gtk.ScrolledWindow.new()
        scrolled_window.set_propagate_natural_height(True)
        scrolled_window.set_min_content_height(200)
        scrolled_window.set_min_content_width(400)
        scrolled_window.set_child(viewport)

        self.main_container.append(scrolled_window)

        self.set_child(self.main_container)

    def _build_ui(self, _: Gtk.Window):
        for app_data in self._installed_apps:
            self.content_container.append(
                AppRowWithCheckbox.build(
                    app_data=app_data,
                    checked=bool(app_data.executable in self._stored_apps)
                )
            )

        connect_button = Gtk.Button.new_with_label(label="Done")
        connect_button.set_name("split-tunneling-app-done-button")
        connect_button.add_css_class("primary")
        connect_button.set_halign(Gtk.Align.END)
        safe_signal_connect(
            connect_button, "clicked", self._on_done_button_clicked
        )
        self.main_container.append(connect_button)

    @GObject.Signal(name="app_selection_completed", arg_types=(object,))
    def app_selection_completed(self, selected_apps: list[AppData]):
        """
        Signal emitted after the user clicks on `Add` button.
        Since GObject can not parse `list[AppData]`, we need to always
        define it as an object.

        Args:
          selected_apps (list[AppData]): a list containing object to be split tunneled.
        """

    def _signal_updated_app_list(self, selected_apps: list[AppData]):
        self.emit("app_selection_completed", selected_apps)

    def _on_done_button_clicked(self, _: Gtk.Button):
        added_apps = []
        child = self.content_container.get_first_child()
        while child:
            if child.checked:
                added_apps.append(child.app_data)
            child = child.get_next_sibling()

        self._signal_updated_app_list(added_apps)
        self.close()

    def _get_first_app_(self) -> AppRowWithCheckbox:
        """Mainly for testing purposes and not for public API.

        Returns:
            AppRowWithCheckbox
        """
        return cast(AppRowWithCheckbox, self.content_container.get_first_child())

    def _click_on_done_button(self):
        """Mainly for testing purposes and not for public API.
        """
        last_child = self.main_container.get_last_child()
        if last_child:
            last_child.emit("clicked")



"""
Changelog window module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import List

from gi.repository import Gtk

from proton.vpn.app.gtk.assets import ASSETS_PATH


class ReleaseNotesDialog(Gtk.Dialog):
    """Dialog that displays release notes for each version,
    in a human friendly way."""

    WIDTH = 450
    HEIGHT = 500
    TITLE = "Release notes"
    RELEASE_NOTES = str(ASSETS_PATH / "release_notes.md")

    def __init__(self):
        super().__init__()
        self.set_default_size(self.WIDTH, self.HEIGHT)
        self.set_title(self.TITLE)
        self.set_modal(True)

        self._content_area = self.get_content_area()

        self.build()

    def build(self):
        """Build the release notes UI."""
        collection = ReleaseNotesCollection()
        collection.create_list(self.RELEASE_NOTES)

        viewport = Gtk.Viewport()
        viewport.set_child(collection)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_propagate_natural_height(True)
        scrolled_window.set_child(viewport)

        self._content_area.append(scrolled_window)


class ReleaseNotesCollection(Gtk.Box):
    """Contains a collection of `ReleaseNote` objects."""
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.set_spacing(25)
        self.add_css_class("release-notes-collection")
        self._release_notes = []

    @property
    def release_notes(self) -> List[ReleaseNote]:
        """Returns all release notes that were added."""
        return self._release_notes

    def create_list(self, filepath: str):
        """Creates the release note based on the path provided by
        `filepath`. This does follow certain patterns:

        ##: Is a header, usually should be used only versions.
        -: Is a bullet point. These bullet points usually belong under a certain
            version, thus these are never by themselves.
        \\n: Means that the bullet points for a specific version have ended and probably
            either a new version will be coming or the file has no more content.
        """
        with open(file=filepath, mode="r", encoding="utf-8") as file:
            lines = file.readlines()
            try:
                last_line = lines[-1]
            except IndexError as excp:
                raise RuntimeError("Release notes file is empty") from excp

        new_entry = ReleaseNote()

        for line in lines:
            self._ensure_log_line_is_valid(line)

            if self.is_last_line_from_file(last_line, line):
                new_entry.add_bullet_point(self.sanitize_log(line))
                new_entry = self._store_and_generate_new_log_entry(new_entry)
            elif self.is_title(line):
                new_entry.add_title(self.sanitize_log(line))
            elif self.is_bullet_point(line):
                new_entry.add_bullet_point(self.sanitize_log(line))
            else:
                new_entry = self._store_and_generate_new_log_entry(new_entry)

    def _ensure_log_line_is_valid(self, line: str):
        if not (line.startswith("##") or line.startswith("-") or line.startswith("\n")):
            raise RuntimeError(
                f"Invalid log line '{line}'.\n"
                "Each log line should start by '##' for headers, "
                "'-' for bullet points or "
                "'\\n' to finish a entry log."
            )

    def is_title(self, current_line: str) -> bool:
        """Returns if `current_line` is a title or not."""
        return current_line.startswith("##")

    def is_bullet_point(self, current_line: str) -> bool:
        """Returns if `current_line` is a bullet point or not."""
        return current_line.startswith("-")

    def is_last_line_from_file(self, last_line: str, current_line: str) -> bool:
        """Returns of the laste line is indeed the last line from the file.

        The reason this comparison is done this way is mainly because if we would
        do a comparison between two strigs (`current_line` and `last_line`) and there
        was another line identical to the last line, then this method would return
        `True` two times, thus messing up the UI. This way we compare against the object
        ID in memory, which is sure to be unique.
        """
        return id(last_line) == id(current_line)

    def sanitize_log(self, line: str) -> str:
        """Sanitizes the output from undesired characters to be displayed."""
        return line.replace("#", "").replace("\n", "").lstrip()

    def _store_and_generate_new_log_entry(self, current_entry):
        self._release_notes.append(current_entry)
        self.append(current_entry)
        return ReleaseNote()


class ReleaseNote(Gtk.Box):
    """Contains all information regarding a release note.

    A release note usually contains a `ReleaseNoteTitle` and a
    collection of `ReleaseNoteBulletPoint` that are coupled to it.
    """
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self._title = None
        self._bullet_points = []

    @property
    def title(self) -> str:
        """Returns the title of this release note."""
        return self._title.get_label()

    @property
    def bullet_points(self) -> List[str]:
        """Returns the bullet points of this release note."""
        return [bullet_point.get_label() for bullet_point in self._bullet_points]

    def add_title(self, title: str):
        """Adds the title to the current `ReleaseNote` object."""
        self._title = Gtk.Label(label=title)
        self._title.set_halign(Gtk.Align.START)
        self._title.set_use_markup(True)
        self._title.add_css_class("heading")

        self.append(self._title)

    def add_bullet_point(self, bullet_point: str):
        """Adds the bullet point to the collection of the current `ReleaseNote` object."""
        bullet_point_label = Gtk.Label(label=bullet_point)
        bullet_point_label.set_halign(Gtk.Align.FILL)
        bullet_point_label.set_wrap(True)
        bullet_point_label.set_max_width_chars(1)
        bullet_point_label.set_property("xalign", 0)
        bullet_point_label.set_property("selectable", True)

        self._bullet_points.append(bullet_point_label)

        self.append(bullet_point_label)



"""
This module contains utility tools.

Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import re
from importlib.metadata import distributions
from typing import Callable
from gi.repository import Gtk

from proton.vpn import logging
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect

# See: https://docs.gtk.org/gtk4/migrating-3to4.html#set-a-proper-application-id  # pylint: disable=line-too-long # noqa: E501
APPLICATION_ID = "proton.vpn.app.gtk"

logger = logging.getLogger(__name__)


def log_proton_package_versions():
    """Logs the versions of all installed proton-* python packages."""
    seen = set()
    packages = []
    for dist in distributions():
        name = dist.metadata["Name"]
        if name and re.match(r"proton", name, re.IGNORECASE) and name not in seen:
            seen.add(name)
            packages.append(f"{name}=={dist.metadata['Version']}")
    logger.info("Proton python packages:\n" + "\n".join(sorted(packages)))


def connect_once(widget: Gtk.Widget, signal: str, callback: Callable) -> int:
    """Subscribe to *widget*'s *signal* once. The handler is disconnected
    after the signal fires for the first time, or after the callback's
    bound instance is collected — whichever comes first.

    Same callback restrictions as safe_signal_connect:
    bound method (held weakly) or pure free function (no closure
    captures); closure-capturing callables raise ``TypeError``."""
    return safe_signal_connect(widget, signal, callback, once=True)


__all__ = ["connect_once"]
